/*
	SkyMedia linux driver code
*/
 

#ifndef MODULE
	#define MODULE
#endif

#include <linux/module.h>
#include <linux/version.h>

#ifdef MODVERSION
	#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/skbuff.h>

#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <asm/irq.h>

#include <linux/delay.h>

#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/fs.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>

#include "sm2_drv.h"

#if LINUX_VERSION_CODE >= 0x020100
    EXPORT_NO_SYMBOLS;
#else
    #include <linux/bios32.h>
    #define ioremap vremap
    #define iounmap vfree
#endif

#if LINUX_VERSION_CODE <= 0x20139
#  include <linux/if_ether.h>
#  define       net_device_stats enet_statistics
#endif

#if LINUX_VERSION_CODE < 0x020100
#  define __initfunc(__arginit) __arginit
#else
#  include <asm/uaccess.h>
#  include <linux/init.h>
#endif

#ifndef TRUE
    #define TRUE	1
    #define FALSE	0
#endif

#ifndef OK
    #define OK	0
    #define ERROR	-1
#endif

/*#define SM200_DEBUG*/

u32 SMIOBASE;
u32 SMIOSIZE;

#define DATAPORT	0
#define COMMANDPORT	2


/* Set statically or when loading the driver module. */
#ifdef DEBUG
static int debug = 1;
#endif

/* the version of driver */

#ifdef SM200_DEBUG
//static char *version = "SkyMedia-200 linux driver Ver1.0\n";
#endif

#ifndef VERSION_CODE
#define VERSION_CODE(vers, rel, seq) (((vers) << 16) | ((rel) << 8) | (seq))
#endif

#if LINUX_VERSION_CODE < VERSION_CODE(2, 4, 0)
#define DEVICE		device
#else
#define DEVICE		net_device
#endif

static int sm200_open_count = 0;

// ------------------------------------------------------------------
// static functions
// ------------------------------------------------------------------
static int sm200_pci_probe(struct DEVICE *dev);
static int sm200_ether_open(struct DEVICE *dev);
static int sm200_ether_close(struct DEVICE *dev);
static int sm200_ether_ioctl(struct DEVICE *dev, struct ifreq *rq, int cmd);
static int sm200_ether_send_packet(struct sk_buff *skb, struct DEVICE *dev);
static struct net_device_stats *sm200_ether_get_stats(struct DEVICE *dev);
static void sm200_set_ether_multicast_list(struct DEVICE *dev);
static void sm200_ether_interrupt(int irq, void *dev_id, struct pt_regs * regs);

// ------------------------------------------------------------------
// global functions
// ------------------------------------------------------------------
int  PMemoryAllocate(void);
void PMemoryDeAllocate(void);
void sm200_udelay(long cnt);

// ------------------------------------------------------------------
// extern functions
// ------------------------------------------------------------------

extern int sm200_ioctl_sub(sm200_ioctl_t *ioc, char * datap);
extern void sm200_network_release(void);
extern void Initialize_Buffer(void);
extern int PCIInitialize(void);
extern int DMAStart(void);
extern int DMAStop(void);
extern int SetMulticastAddresses(u8 MulticastCount, u8 Mul_address[8][6] );
extern int AVVXP195_IRQ_Hw_Int_Proc(void *dev);
extern int PLX_IRQ_Hw_Int_Proc(void *dev);
extern int ReadMacAddress(void);

int my_str_cpy(char *dest, char *src);


// ------------------------------------------------------------------
/* Global variables */
// ------------------------------------------------------------------

#ifdef SM200
	char sky_if_name[20] = "sm200";

	#define PCI_CHIP_VENDOR_ID    0x11d1
	#define PCI_CHIP_DEVICE_ID    0x01f8
	#define TELEMANN	0x1378
	#define SKYMEDIA	0x3210
#endif
#ifdef SM300
	char sky_if_name[20] = "sm300";
	#define PCI_CHIP_VENDOR_ID    0x11d1
	#define PCI_CHIP_DEVICE_ID    0x01f8
	#define TELEMANN	0x1378
	#define SKYMEDIA	0x3210
#endif
#ifdef SM200D
	#define MAX_CHIP_TYPE  3
	#define MAX_PRODUCT_NO 3
	#define sm200d		0
	#define sm200dtp	1
	#define sm200dpa	2
	
	char sky_if_name[20] = "sm200d";
	char ProductName[20];
	int ProductNo;
	u16 PCI_CHIP_VENDOR_ID[] = { 0x11d1 /* SM200D */, 0x1378/*SM200DTP*/, 0x1378/*SM200DPA*/ };
	u16 PCI_CHIP_DEVICE_ID[] = { 0x01f8 /* SM200D */, 0x4050/*SM200DTP*/, 0x4020/*SM200DPA*/ };
	u16 PRODUCT_VENDOR[] = { 0x1378, 0x1378, 0x1378 };
	u16 PRODUCT_SUBSYSTEM[] = { 0x3240, 0x4050, 0x4020 };
	u8 *ProductNameList[MAX_PRODUCT_NO] = { "SM200D", "SM200DTP", "SM200DPA" };
	u32 SMIOBASE_LIST[] = { 0x80000, 0x0000, 0x0000 };
	u32 SMIOSIZE_LIST[] = { 0x8000, 0x0100, 0x0100 };
	
	u8 *DMAalignedLinear1;
#endif

struct DEVICE *sm200_dev = NULL;

volatile u32	*LinAddress;
volatile u32	*LinVxPAddress;
u8 PORTA_ADDRESS;
u8 PORTB_ADDRESS;
u8 PCI_WR_PTR;
u8 PCI_RD_PTR;
u8 PORTC_ADDRESS;
u8 PORTD_ADDRESS;
u8 INTCSR;
u8 DMA_ALIGN_SIZE;
u32 DMA_ALIGN_MASK;
u16 DMA_BUF_SIZE;
u16 DMA_BUFFER_MASK;

#define MAX_PRIV_LEN	512	

typedef struct sm2_priv {
	struct DEVICE *next;
	u32	irqlock;
	struct net_device_stats stat;
	char io_buf[MAX_PRIV_LEN];	
} sm2_priv;
u8 Mul_address[8][6];
u8 Mul_address_count = 0;

// ------------------------------------------------------------------
/* extern variables */
// ------------------------------------------------------------------
extern u8  OMacAddress[6];
extern u8  MacHeader[14];
extern u32 DMABufferLinear1;
extern u32 DMAPhysicalAddress1;
extern u32 TableLogicalAddress;
extern u32 TablePhysicalAddress;
// ------------------------------------------------------------------


int init_module(void)
{
	int retval;

#ifdef SM200_DEBUG
	printk(KERN_ALERT "Initializing Skymedia card.\n");
	printk(KERN_ALERT "====== Debug Version ======\n");
#endif
	/* We must emit version information. */
	retval = sm200_pci_probe(0);

	if (retval) {
#ifdef SM200_DEBUG
		printk(KERN_NOTICE "skymedia: no (useable) cards found, driver NOT installed(erro=%d).\n", retval);
#endif
		return retval;
	}
#ifdef SM200_DEBUG
	printk(KERN_ALERT "Skymedia card is initialized.\n");
#endif

	return OK;
}

#if LINUX_VERSION_CODE < VERSION_CODE(2, 4, 0)
__initfunc(static int sm200_pci_probe(struct DEVICE *dev))
#else
static int sm200_pci_probe(struct DEVICE *dev)
#endif
{
	
	int pci_result;
	struct pci_dev *pdev = NULL;
	u16 subvendorid, subsysid, pci_command, new_command;
	u_long mem_ptr_phys;
	void *mem_ptr_virt;
	unsigned char pci_bus, pci_func;
	u8 pci_irq;
	ulong pci_ioaddr;
	int pci_index, result;
	int i;
	int chip_type;

	pdev = NULL;
	
	if(!pcibios_present())
	{
#ifdef SM200_DEBUG
		printk(KERN_DEBUG "skymedia: no pci_bios present\n");
#endif
		return -ENODEV;
	}

	for(pci_index=0; pci_index < 0xff; pci_index++)
	{
		for(chip_type = 0; chip_type < MAX_CHIP_TYPE; chip_type++)
		{
			result = pcibios_find_device(PCI_CHIP_VENDOR_ID[chip_type],
						PCI_CHIP_DEVICE_ID[chip_type], pci_index, &pci_bus, &pci_func);

			if(result==PCIBIOS_SUCCESSFUL)
		        break;
		}
		if(result==PCIBIOS_SUCCESSFUL)
	        break;
		
#ifdef SM200_DEBUG
		printk(KERN_DEBUG "skymedia: pcibios_find_device failed:%d\n", pci_index);
#endif
	}

	if (result != PCIBIOS_SUCCESSFUL) {
#ifdef SM200_DEBUG
		printk(KERN_DEBUG "sm200: no card found\n");
#endif
		return -ENODEV;
	}

#if LINUX_VERSION_CODE >= 0x20155
        pdev = pci_find_slot(pci_bus, pci_func);
#if LINUX_VERSION_CODE < VERSION_CODE(2, 4, 0)
        pci_ioaddr = pdev->base_address[0] & ~3;a
#else
	pci_ioaddr = pci_resource_start(pdev, 0);
#endif
        pci_irq = pdev->irq;
#else
        pcibios_read_config_byte(pci_bus, pci_func,
                                  PCI_INTERRUPT_LINE, &pci_irq);
        pcibios_read_config_dword(pci_bus, pci_func,
                                 PCI_BASE_ADDRESS_0, (int*)&pci_ioaddr);
        pci_ioaddr = pci_ioaddr & ~3;;
#endif

        pci_result = pcibios_read_config_word(pci_bus, pci_func, PCI_SUBSYSTEM_VENDOR_ID, &subvendorid);
        pci_result = pcibios_read_config_word(pci_bus, pci_func, PCI_SUBSYSTEM_ID, &subsysid);
	

/* Let's find what produdct has been installed */

		for(i=0; i< MAX_PRODUCT_NO; i++)
		{
			if ( ((subvendorid == PRODUCT_VENDOR[i]) && (subsysid == PRODUCT_SUBSYSTEM[i]))
				|| ((subsysid == PRODUCT_VENDOR[i]) && (subvendorid == PRODUCT_SUBSYSTEM[i])) )
			{
				/* We have found the card */
				ProductNo = i;
				break;
			}
		}
		if(i == MAX_PRODUCT_NO)
		{
#ifdef SM200_DEBUG			
			printk(KERN_NOTICE "skymedia :VENDOR ID:%x or DEVICE ID:%x wrong !!!\n", subvendorid, subsysid);
#endif
			return -ENODEV;
		}

		/* set various parameters for each card type */
		strcpy(ProductName, ProductNameList[ProductNo]);
		SMIOBASE = SMIOBASE_LIST[ProductNo];
		SMIOSIZE = SMIOSIZE_LIST[ProductNo];

#ifdef	SM200_DEBUG
		printk(KERN_NOTICE "skymedia : Detected Card = %s\n", ProductName);
#endif
		if(ProductNo == sm200dtp || ProductNo == sm200dpa)
		{
			
			PORTA_ADDRESS = (0x50/4);	// data port	: Mailbox Register4
			PORTB_ADDRESS = (0x54/4);	// control port	: Mailbox Register5
			PCI_WR_PTR	 =	(0x58/4);	// Mailbox register 6 (DMA write ptr)
			PCI_RD_PTR	 =	(0x5C/4);	// Mailbox register 7 (DMA read ptr)
			PORTC_ADDRESS =	(0x60/4);	// PCI -> LOCAL		: PCI-to-Local Doorbell
			PORTD_ADDRESS =	(0x64/4);	// LOCAL->PCI intr	: Local-to-PCI Doorbell
			INTCSR		 =	(0x68/4);
			DMA_ALIGN_SIZE =	0x4;
			DMA_ALIGN_MASK =	0xFFFFFFFC;
			DMA_BUF_SIZE =	0x8000;	// 32KB
			DMA_BUFFER_MASK =	0x7FFF; 
		}
		else
		{
			PORTA_ADDRESS =  0x00;
			PORTB_ADDRESS =  0x02;
		}

		
#if 0
        printk(KERN_DEBUG "Subvendor ID : %x\n", subvendorid);
        printk(KERN_DEBUG "Vendor : %x\n", pdev->vendor);
        printk(KERN_DEBUG "Device : %x\n", pdev->device);
        printk(KERN_DEBUG "Func : %x\n", pdev->devfn);
        printk(KERN_DEBUG "Class : %x\n", pdev->class);
        printk(KERN_DEBUG "IRQ : %x\n", pdev->irq);
        printk(KERN_DEBUG "Base address 0 : 0x%8x\n", (int)pci_ioaddr);
        printk(KERN_DEBUG "Initialized Skymedia.\n");
#endif

        // We found Skymedia here.
        // Bus Master Setup here.

        pcibios_read_config_byte(pci_bus, pci_func, PCI_COMMAND, (u8 *)&pci_command);
        new_command = pci_command | PCI_COMMAND_MASTER;
        pcibios_write_config_byte(pci_bus, pci_func, PCI_COMMAND, (u8)new_command);

        // remap physcal address
        mem_ptr_phys = (u_long)pci_ioaddr;
        if((mem_ptr_virt=ioremap((u_long)mem_ptr_phys,SMIOSIZE))==NULL)
        {
#ifdef SM200_DEBUG
                printk(KERN_DEBUG "sm200: ioremap for 0x%x bytes high PCI memory at 0x%lx failed\n",
                        SMIOSIZE, (u_long)mem_ptr_phys );
#endif
                LinVxPAddress = (u32 *)mem_ptr_phys;
        }
        else
        {
#ifdef SM200_DEBUG
                printk(KERN_DEBUG "sm200: remapped 0x%x bytes high PCI memory at 0x%lx to 0x%lx.\n",
                        SMIOSIZE, (u_long)mem_ptr_phys, (u_long)mem_ptr_virt);
#endif
                LinVxPAddress = (u32 *)mem_ptr_virt;
        }

        mem_ptr_phys = (u_long)pci_ioaddr + SMIOBASE;
        if((mem_ptr_virt=ioremap((u_long)mem_ptr_phys,SMIOSIZE))==NULL)
        {
#ifdef SM200_DEBUG
                printk(KERN_DEBUG "sm200: ioremap for 0x%x bytes high PCI memory at 0x%lx failed\n",
                        SMIOSIZE, (u_long)mem_ptr_phys );
#endif
                LinAddress = (u32 *)mem_ptr_phys;
        }
        else
        {
#ifdef SM200_DEBUG
                printk(KERN_DEBUG "sm200: remapped 0x%x bytes high PCI memory at 0x%lx to 0x%lx.\n",
                        SMIOSIZE, (u_long)mem_ptr_phys, (u_long)mem_ptr_virt);
#endif
                LinAddress = (u32 *)mem_ptr_virt;
        }


	// Initialize PCI
	pci_result = PCIInitialize();
	if(pci_result)
		return ERROR;

	// Allocate memory for DMA
	pci_result = PMemoryAllocate();
	if(pci_result)
		return ERROR;

	// Start DMA
	pci_result = DMAStart();
	if(pci_result)
		return ERROR;

   	Initialize_Buffer();
	if (NULL == dev) {
		dev = init_etherdev(NULL, sizeof(sm2_priv));
	}

	dev->base_addr = pci_ioaddr + SMIOBASE;
	dev->irq = pci_irq;

	my_str_cpy(dev->name,sky_if_name);
	
	dev->open = &sm200_ether_open;
	dev->stop = &sm200_ether_close;
	dev->hard_start_xmit = &sm200_ether_send_packet;
	dev->get_stats = &sm200_ether_get_stats;
	dev->do_ioctl = &sm200_ether_ioctl;
	dev->set_multicast_list = &sm200_set_ether_multicast_list;

#if defined(SM200D) || defined(SM300)
	/* read MAC address from the card and set the OMacAddress[] */
	ReadMacAddress(); 
	((u8*)dev->dev_addr)[0] = OMacAddress[0];
	((u8*)dev->dev_addr)[1] = OMacAddress[1];
	((u8*)dev->dev_addr)[2] = OMacAddress[2];
	((u8*)dev->dev_addr)[3] = OMacAddress[3];
	((u8*)dev->dev_addr)[4] = OMacAddress[4];
	((u8*)dev->dev_addr)[5] = OMacAddress[5];
#endif
	sm200_dev = dev;
	pci_result = request_irq(dev->irq, sm200_ether_interrupt, SA_SHIRQ, dev->name, dev);

	return pci_result;
}

void cleanup_module(void)
{
	free_irq(sm200_dev->irq, sm200_dev);
	sm200_network_release();


	DMAStop();
	PMemoryDeAllocate();
	sm200_ether_close(sm200_dev);
	unregister_netdev(sm200_dev);
	iounmap((u32*)LinVxPAddress);
	iounmap((u32*)LinAddress);
#ifdef SM200_DEBUG
	printk(KERN_DEBUG "Skymedia 200 unloaded.\n");
#endif
}

static int sm200_ether_open(struct DEVICE *dev)
{
	struct sm2_priv *sm2_local = (struct sm2_priv *) dev->priv;
       
	if(sm200_open_count)
		return -EBUSY;  /* already open */

	sm200_open_count++;
#if LINUX_VERSION_CODE < VERSION_CODE(2, 4, 0)
	dev->start = 1;
#else
	netif_start_queue(dev);
#endif
	sm2_local->irqlock = 0;
	MOD_INC_USE_COUNT;
	return 0;
}

static int sm200_ether_close(struct DEVICE *dev)
{
	sm200_open_count--;
#if LINUX_VERSION_CODE < VERSION_CODE(2, 4, 0)
	dev->start = 0;
#else
	netif_stop_queue(dev);
#endif
	MOD_DEC_USE_COUNT;
	return 0;
}
static int sm200_ether_ioctl(struct DEVICE *dev, struct ifreq *rq, int cmd)
{

	int bResult=0;
	sm200_ioctl_t ioc;
	char *userp;
	char *kernelp = 0; 
	struct sm2_priv *priv = (struct sm2_priv *) dev->priv;

#if LINUX_VERSION_CODE >= 0x020100
	copy_from_user(&ioc, rq->ifr_data, sizeof(sm200_ioctl_t));
#else
	if((bResult = verify_area(VERIFY_READ, rq->ifr_data, sizeof(sm200_ioctl_t)))!=0)
	{
#ifdef	SM200_DEBUG
		printk(KERN_NOTICE "skymedia: verfiy_area failed.\n"); 
#endif
		return bResult;
	}
	memcpy_fromfs(&ioc, rq->ifr_data, sizeof(sm200_ioctl_t) );
#endif
	if(ioc.len > MAX_PRIV_LEN)
	{
#ifdef	SM200_DEBUG
		printk(KERN_NOTICE "sm200 : cmd %x want too large kerne memory (%d bytes).\n",ioc.cmd, ioc.len );
#endif
		return -1;	// error
	}
	kernelp = priv->io_buf;
        userp = ioc.dt.datap;

#if LINUX_VERSION_CODE >= 0x020100
	if(ioc.len)
		copy_from_user(kernelp, userp, ioc.len );
	bResult = sm200_ioctl_sub(&ioc, kernelp);
	if(bResult != 0)
	{
	#ifdef	SM200_DEBUG
		printk(KERN_NOTICE "sm200 : ioctl_sub error \n");
		printk(KERN_NOTICE "  ==== cmd:%x, len:%d, err:%x\n", ioc.cmd, ioc.len, bResult);
	#endif
		return bResult;
	}

	if(ioc.len)
   		copy_to_user(userp, kernelp, ioc.len);
#else
	if((bResult = verify_area(VERIFY_WRITE, userp, ioc.len))!=0)
	{
	#ifdef	SM200_DEBUG
		printk(KERN_NOTICE "skymedia: verfiy_area failed.\n"); 
	#endif
		return bResult;
	}
	if(ioc.len)
		memcpy_fromfs(kernelp, userp, ioc.len );

	bResult = sm200_ioctl_sub(&ioc, kernelp);
	if(bResult != 0)
	{
	#ifdef	SM200_DEBUG
		printk(KERN_NOTICE "sm200 : ioctl_sub error \n");
		printk(KERN_NOTICE "  ==== cmd:%x, len:%d, err:%x\n", ioc.cmd, ioc.len, bResult);
	#endif
		return bResult;
	}
	if(ioc.len)
		memcpy_tofs(userp, kernelp, ioc.len);
#endif
	return bResult;

}

static int sm200_ether_send_packet(struct sk_buff *skb, struct DEVICE *dev)
{
	return 0;
}

static struct net_device_stats *sm200_ether_get_stats(struct DEVICE *dev)
{
	struct sm2_priv *sm2_local = (struct sm2_priv *) dev->priv;
    
	/* If the card is stopped, just return the present stats. */
#if LINUX_VERSION_CODE < VERSION_CODE(2,4,0)
	if (dev->start == 0) 
		return &sm2_local->stat;
#endif

	return &sm2_local->stat;
}

static void sm200_set_ether_multicast_list(struct DEVICE *dev)
{
//	struct sm2_priv *sm2_local = (struct sm2_priv *) dev->priv;
	unsigned long flags;
	struct dev_mc_list *dmi;
	u8 mcount = 0;
	int j;

	for(dmi=dev->mc_list; mcount < 8 && dmi!=NULL; dmi=dmi->next) {
		for( j = 0; j < 6; j++) {
			Mul_address[mcount][j] = dmi->dmi_addr[j];
		}

#ifdef SM200_DEBUG
		printk(KERN_NOTICE "Multicast_addr[%d] = %2x:%2x:%2x:%2x:%2x:%2x\n",
			mcount,
			Mul_address[mcount][0],
			Mul_address[mcount][1],
			Mul_address[mcount][2],
			Mul_address[mcount][3],
			Mul_address[mcount][4],
			Mul_address[mcount][5]);
#endif
		mcount++;
	}
		
	save_flags(flags);
//	cli();
	Mul_address_count = mcount;
	SetMulticastAddresses(mcount, Mul_address);
	restore_flags(flags);

	return;
}

static void sm200_ether_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
	struct DEVICE *dev = dev_id;
#if LINUX_VERSION_CODE < VERSION_CODE(2, 4, 0)
	struct sm2_priv *sm2_local = (struct sm2_priv *) dev->priv;
#endif    
	if (dev == NULL) 
	{
		return;
	}
    
#if LINUX_VERSION_CODE < VERSION_CODE(2, 4, 0)
	if (dev->interrupt || sm2_local->irqlock) 
	{
#if 1 /* This might just be an interrupt for a PCI device sharing this line */
		/* The "irqlock" check is only for testing. */
#ifdef SM200_DEBUG
		printk(KERN_DEBUG "Reentering the interrupt handler!\n");
#endif
#endif
		return;
	}
    
	dev->interrupt = 1;
#endif

	if(ProductNo == sm200dtp || ProductNo == sm200dpa )
	{
		PLX_IRQ_Hw_Int_Proc((void *)dev);
	}
	else
	{
		AVVXP195_IRQ_Hw_Int_Proc((void *)dev);
	}		
#if LINUX_VERSION_CODE < VERSION_CODE(2, 4, 0)
	dev->interrupt = 0;
#endif
	return;
}

void sm200_ether_receive(void *device, u16 pkt_len, u8 *pkt_ptr, u32 multicast)
{
	struct DEVICE *dev = (struct DEVICE *)device;
	struct sm2_priv *sm2_local = (struct sm2_priv *) dev->priv;
	struct sk_buff *skb;
	u8 *temp;
	int i;

	if(pkt_len < 32  ||  pkt_len > 1504) 
	{
		sm2_local->stat.rx_errors++;
		sm2_local->stat.rx_length_errors++;
	}
	else 
	{
		skb=alloc_skb(pkt_len+14+2, GFP_ATOMIC);
		if(skb == NULL) 
		{
#ifdef SM200_DEBUG
			printk(KERN_DEBUG "%s: Couldn't allocate a sk_buff of size %d.\n",
					   dev->name, pkt_len);
#endif
			sm2_local->stat.rx_dropped++;
			return;
		}
		else
		{
			skb->dev = dev;
			skb_reserve(skb,2);

			temp = skb->data;
			for(i=0; i<6; i++) 
			{
				*(u8 *)(temp+i) = MacHeader[5-i%6];
			}
			for( i = 0; i < 6; i++ ) {
				*(u8 *)(temp+i+6) = skb->dev->dev_addr[i];
			}
			*(u8 *)(temp+12)=0x08; 
			*(u8 *)(temp+13)=0x00;
			skb->len+=14;
			skb->tail+=14;

			memcpy(skb_put(skb,pkt_len), pkt_ptr, pkt_len);

			skb->protocol=eth_type_trans(skb,skb->dev);

			if(multicast) 
			{
				skb->pkt_type = PACKET_MULTICAST;
			}
			else
				skb->pkt_type = PACKET_HOST;

			netif_rx(skb);
			sm2_local->stat.rx_packets++;
#if LINUX_VERSION_CODE >= 0x020100
			sm2_local->stat.rx_bytes += pkt_len;
#endif
		}
	} 
}


int PMemoryAllocate(void)
{


	u32 phy_aligned_start;

	if(ProductNo == sm200dtp || ProductNo == sm200dpa)
	{
		DMABufferLinear1 = (u32)kmalloc(PAGE_SIZE*(16+8), GFP_KERNEL | GFP_DMA);
	   	if(!DMABufferLinear1)
	   	{
		#ifdef SM200_DEBUG
		    printk(KERN_NOTICE "skymedia : dma page allocation error\n");
		#endif
		    return ERROR;
	   	}
		DMAPhysicalAddress1 = virt_to_phys((volatile void*)DMABufferLinear1);
	
		/* it is for PLX9054.
		   it needs 64KB aligned physical start address. */
		
		phy_aligned_start = ((u32)DMAPhysicalAddress1 + 0xffff) & 0xffff0000;
	
		DMAPhysicalAddress1 = phy_aligned_start;
		DMAalignedLinear1 = (u8*)phys_to_virt(phy_aligned_start);
	}	
	else
	{

		DMABufferLinear1 = (u32)kmalloc(8*PAGE_SIZE, GFP_KERNEL | GFP_DMA);
		DMAalignedLinear1 = (u8*)DMABufferLinear1;
		
	   	if(!DMABufferLinear1)
	   	{
		#ifdef SM200_DEBUG
		    printk(KERN_NOTICE "skymedia : dma page allocation error\n");
		#endif
		    return ERROR;
	   	}
		DMAPhysicalAddress1 = virt_to_phys((volatile void*)DMABufferLinear1);

		TableLogicalAddress = (u32)kmalloc(1*PAGE_SIZE, GFP_KERNEL | GFP_DMA);
	   	if(!TableLogicalAddress)
	   	{
		#ifdef SM200_DEBUG
		    printk(KERN_NOTICE "skymedia : dma page allocation error\n");
		#endif
		    return ERROR;
	   	}
		TablePhysicalAddress = virt_to_phys((volatile void*)TableLogicalAddress);

	}

   	return OK;
}

void PMemoryDeAllocate(void)
{
	if(ProductNo == sm200dtp || ProductNo == sm200dpa)
	{
	    kfree((void *)DMABufferLinear1);
	    kfree((void *)TableLogicalAddress);
	}
	else
	{
	    kfree((void *)TableLogicalAddress);
	}
    return;
}

void sm200_udelay(long cnt)
{
  udelay(cnt*10);
}

int my_str_cpy(char *dest, char *src)
{
	while(*src != 0)
	{
		*dest++ = *src++;
	}
	*dest = 0;
	
	return 0;
}

